#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/*
 * This is the final version of the getinfo program
 *
 * It will read a customer text file (cust.txt) and index file (cust.ndx)
 * It allows the user to add, delete, and edit entries.
 * The user can also find specific entries
 * Finally, the user can save their changes to cust.txt
 *
 * In addition, the program makes sure that duplicate records don't exist
 * for the same customer ID.  
 *
 * The user can also new and changed records are added to the end of the
 * file.  Old records are marked as deleted.  The user may compress the
 * file, removing deleted records.
 */

#include "getinfo.h"

/* function prototypes are in getinfo.h */

main()
{
    char buf[BUFSIZ];
    int ncust=0;    /* number of customers */
    int i;          /* loop variable */
    FILE *fp;       /* file for saving records */
    FILE *keyfp;    /* index file */
    struct index index[MAXENT];   /* index for file */
    struct cust *cust;  /* temp work record */
    int opt;        /* menu option */
    int ent;        /* current entry position in index array */

    /*
     * read index file into memory
     * SInce this is a binary file, we must specify binary
     */
    if ((keyfp = fopen("cust.ndx", "rb")) == NULL)
    { /* no index file - must be new database */
        fprintf(stderr, "Can't read index file - starting with empty file\n");
        index[0].id = -1;   /* signal empty file */
        index[0].loc = -1L;
        /* open customer file as new file for update */
        fp = fopen("cust.txt", "w+");
    }
    else
    { /* index file exists - read it in and determine how many customers */
        fread(index, sizeof(struct index), MAXENT, keyfp); /* read index file */
        fclose(keyfp);  /* close index file */
        /* find end of index table */
        for (ncust=0; ncust < MAXENT && index[ncust].id > 0; ncust++);
        /* open customer file as existing file for update */
        fp = fopen("cust.txt", "r+");
    }

    /*
     * We have opened the index file if it existed
     * We have tried to open the customer file for update (existing or new)
     * If we couldn't open the customer file, we can't proceed
     */
    if (fp == NULL)
    { /* serious problems - can't open text file */
    	fprintf(stderr, "Can't open customer file for update!\n");
	    exit(1);    /* terminate program - fatal error */
    }

    /*
     * we now have a list of the existing customers in memory
     *
     * loop getting commands from user.  Process commands until done
     */
    do
    { /* display menu, get option, process option ... until done */
        /* display menu */
        display_menu();
        printf("\nEnter option, or '?' for help: ");
        gets(buf);  /* get option from user */
        if (buf[0] == '\0') continue;    /* nothing entered */
        if (buf[0] == 'q') /* undocumented way to quit */
        { /* user wants to quit - make sure */
        	printf("Are you sure you want to quit? (y or n): ");
         gets(buf);
         if (buf[0] == 'y') break;	/* quit loop */
        }
        if (buf[0] == '?')
        { /* user asked for help */
        	do_help();	/* display help message */
         continue;	/* get next option */
        }
        opt = atoi(buf);    /* convert to number */

        /*
         * User has entered an option.  Process it
         */
        switch(opt)
        {
        case 1: /* add customer */
            if (ncust > MAXENT-1)
            { /* table full - can't add */
                printf("Customer File is Full! Max. size is %d\n", MAXENT);
                printf("Delete some records, or increase size of MAXENT\n");
                continue;
            }
            cust = get_cust();  /* get customer record from user */
            if (cust == NULL) continue;    /* user cancelled add */
            /* make sure same customer ID doesn't exist */
            ent = find_cust(index, cust->id);
            if (ent >= 0)
            { /* duplicate found */
            	printf("Customer %d already exists!  Add cancelled!\n");
                free_cust(cust); /* free up customer record */
                continue;   /* get another option */
            }
            /* ready to add customer - find where to add record */
            for (i=0; i < ncust; i++)
            { /* as soon as we find an ID bigger than new ID, quit */
                if (index[i].id > cust->id) break;
            }
            ent = i;    /* where to insert new record */
            /*
             * now push all entries past insertion slot down one position
             * we have to start at the bottom and move up
             * otherwise, we would copy entry 'ent' to the rest of the array
             */
            for (i=ncust-1; i >= ent; i--)
            { /* move this entry up one position */
                index[i+1].id = index[i].id;
                index[i+1].loc = index[i].loc;
            }
            ncust++;	/* one more customer */
            /* mark new last entry in index table */
            index[ncust].id = -1;
            index[ncust].loc = -1L;

            /* add new entry to index and file */
            index[ent].id = cust->id;
            fseek(fp, 0, SEEK_END); /* go to end of file */
            index[ent].loc = ftell(fp); /* where we will insert record */
            write_cust(index+ent, cust, fp);   /* do the actual write */
            printf("Customer record added at location %ld\n", index[ent].loc);
            break;

        case 2: /* delete customer */
            /* get customer ID and find it */
            printf("Enter customer ID to find: ");
            gets(buf);
            ent = find_cust(index, atoi(buf));
            if (ent < 0)
            {
                fprintf(stderr, "Customer record not found\n");
                continue;
            }
            /* found customer record - now delete it */
            cust = read_cust(index+ent, fp);  /* read record */
            display_cust(cust);   /* display customer record */
            free_cust(cust);    /* free memory */
            printf("Do you want to delete this customer? (y or n): ");
            gets(buf);
            if (buf[0] != 'y') continue;    /* don't delete */
            
            /*
             * we are to delete the current record
             * remove entry from index array
             * by shifting all the following entries up one position
             * and decrementing the number of customers
             */
            for (i=ent; i < ncust-1; i++)
            { /* shift index entry up one position */
                index[i].id = index[i+1].id;
                index[i].loc = index[i+1].loc;
            }
            ncust--;    /* one less customer */
            index[ncust].id = -1;	/* mark end of index */
            index[ncust].loc = -1L;	/* mark new end of index */
            break;

        case 3: /* edit customer */
        	printf("Enter customer ID to edit: ");
         gets(buf);
         ent = find_cust(index, atoi(buf));	/* find customer entry */
         if (ent < 0)
         {
         	fprintf(stderr, "Can't find customer record\n");
            continue;
         }
         cust = read_cust(index+ent, fp);	/* read customer record */
         if (cust == NULL)
         { /* error reading file */
         	fprintf(stderr, "Error reading customer record\n");
            continue;
         }
         edit_cust(cust);	/* let user edit the customer */
         display_cust(cust);
         printf("Update customer record? (y or n): ");
         gets(buf);
         /* update the record unless the user says no */
         if (buf[0] != 'n') write_cust(index+ent, cust, fp);
         free_cust(cust);   /* free up customer */
         break;

        case 4: /* find customer */
        	printf("Enter Customer ID to find: ");
         gets(buf);
         ent = find_cust(index, atoi(buf));
         if (ent < 0)
         {
         	printf("Customer record not found!\n");
            continue;
         }
         cust = read_cust(index+ent, fp);	/* read record */
         display_cust(cust);	/* and display record */
         free_cust(cust);       /* free up memory */
         break;

        case 5: /* list customers */
        	list_cust(index, fp);	/* list customer records */
         break;

        case 6: /* search by last name */
            printf("Enter last name: ");
            gets(buf);
            search_lname(index, buf, fp);  /* search for all records with this lname */
            break;

        case 90: /* compress file */
        	printf("Compressing customer file ...\n");
        	compress_file(index, fp);
         printf("Customer File compressed\n");
         /* we had to close the file to do the renaming */
         fp = fopen("cust.txt","r+");	/* open for existing file */
         if (fp == NULL)
         { /* problems */
         	fprintf(stderr, "Fatal error: Can't re-open customer file\n");
            exit(1);
         }
         break;

        case 99: /* quit */
            printf("Are you sure you want to quit? (y or n): ");
            gets(buf);
            if (buf[0] != 'y') opt = 0; /* don't quit */
            break;
        default:	/* invalid option */
        	fprintf(stderr, "Invalid menu option (%s)\n", buf);
        }
    } while (opt != 99);    /* loop until done */

    if ((fp = fopen("cust.ndx", "wb")) == NULL)
    {
        fprintf(stderr, "Can't open index file\n");
        exit(3);
    }
    /* file open - write entire array */
    index[ncust].id = -1;		/* signal end of array */
    index[ncust].loc = -1L;   /* by putting illegal values at end */
    fwrite(index, sizeof(struct index), MAXENT, fp);  /* write entire table */
    fclose(fp);     /* close the file */


    printf("Press ENTER to terminate program: "); /* display another message */
	getchar(); 						/* wait for the user to press ENTER */
    return 0;
}

/*
 * Display menu for user
 */
void display_menu()
{
	printf("\nOptions are:\n");
    printf(" 1. Add Customer\n");
    printf(" 2. Delete Customer\n");
    printf(" 3. Edit Customer\n");
    printf(" 4. Find Customer\n");
    printf(" 5. List Customers\n");
    printf(" 6. Search by Last Name\n");
    printf("90. Compress File\n");
    printf("99. Quit\n");
}

void display_cust(struct cust *cust)
{
    printf("Customer %d: %s %s, bal %.2f\n", cust->id, cust->fname,
    		cust->lname, cust->bal);
}

/*
 * This function reads the specified entry from the file
 * Before this routine is called, the file should be
 * positioned to the start of the record
 *
 * If the entry is valid, it is returned.
 * If any problems are found, NULL is returned
 *
 * This file is structured differently than previous files
 * Each entry is on a single line
 * The individual fields are separated by vertical bars (|)
 *
 */
struct cust *read_cust(struct index *ndx, FILE *fp)
{ /* read entry from file into cust table */
    char buf[BUFSIZ], *p, *q;
    struct cust *cust;

    /* position to record in file */
    fseek(fp, ndx->loc, SEEK_SET);

    /* read line from file */
    if (read_line(buf, fp)) return NULL;    /* error reading */

    /* record read - now allocate record and parse record into fields */
    if ((cust = alloc_cust()) == NULL)
    {
    	fprintf(stderr, "Can't allocate memory for cust rec\n");
      return NULL;
    }

    /* first field is customer ID */
    for (p=buf; isspace(*p); p++);	/* find non-blank */
    for (q=p; *q != '\0' && *q != '|'; q++);	/* find end of field */
    if (*q != '\0') *q++ = '\0';	/* end of field */
    cust->id = atoi(p);

    /*
     * We have to make copies of any string fields
     * since the character array buf only exists when
     * this routine is active
     */
    /* next is last name */
    for (p=q; *q != '\0' && *q != '|'; q++);	/* find end of field */
    if (*q != '\0') *q++ = '\0';	/* terminate field */
    cust->lname = strdup(p);	/* make copy of string for last name */

    /* next field is first name */
    for (p=q; *q != '\0' && *q != '|'; q++);	/* find end of field */
    if (*q != '\0') *q++ = '\0';	/* terminate field */
    cust->fname = strdup(p);	/* copy for first name */

    /* last field is balance */
    cust->bal = atof(q);	/* no need to parse any further */

    return cust;
}
/*
 * This function allocates memory for a customer record
 * and initializes its fields
 *
 * In object oriented languages, this is called a constructor
 *
 * When allocating memory, it is very important that you free memory
 * when you are done using it.  If not, then 'holes' start appearing
 * in memory.  The program will still run correctly, but each time it
 * has less and less memory to work with.  Eventually, the program
 * will run out of memory.  This is called a "memory leak"
 */
struct cust *alloc_cust()
{
	struct cust *cust;

	cust = (struct cust *)malloc(sizeof (struct cust));
   if (cust != NULL)
   { /* allocated - initialize fields */
   	cust->id = 0;
      cust->lname = NULL;
      cust->fname = NULL;
      cust->bal = 0.0;
   }
   return cust;
}

/*
 * This function frees up the memory associated with the customer rec
 *
 * In object-oriented languages, this is called a destructor
 *
 * Since our structure contains other memory that has been allocated,
 * it is important that we free those fields as well.
 *
 * See the comments for the function alloc_cust about memory leaks
 */
void free_cust(struct cust *cust)
{
    /* 
     * we test the fields for NULL to determine if memory has been allocated
     *
     * When alloc_cust is called, these fields are initialized to NULL
     * If we call free_cust right after alloc_cust, then the name fields
     * wouldn't have any allocated memory associated with them
     *
     * This won't happen the way the program is currently written,
     * but it is a good idea to check just in case.
     */
 	if (cust->lname != NULL) free(cust->lname);	/* free last name */
	if (cust->fname != NULL) free(cust->fname);	/* free first name */
   	free(cust);				/* hope nobody uses this again */
}

/*
 * Read line from file
 * strip end of line sequence (CR-LF)
 *
 * If read fails, return non-zero; otherwise return zero
 */
int read_line(char buf[], FILE *fp)
{ /* read line from file and strip 'end-of-line' sequence */

    /*
     * read next line from file
     *
     * If end of file encountered, return 1
     * otherwise, return 0
     */
    if (fgets(buf, BUFSIZ, fp)==NULL) return 1;

    /* find end of line - CR, LF, or binary zero */
    while (*buf != '\0' && *buf != '\r' && *buf != '\n') buf++;
	 *buf = '\0';	/* terminate string */

    return 0;	/* normal return */
}

/*
 * This function writes out the given customer record to the file
 *
 * Records are always written at the end of the file,
 * and the location is updated
 *
 * The record is written as a single line with vertical bars between fields
 */
void write_cust(struct index *ndx, struct cust *cust, FILE *fp)
{ /* write customer record out to file */

	/* go to end of file and write record */
   fseek(fp, 0, SEEK_END);	/* go to end of file */
   ndx->loc = ftell(fp);	/* save location for record */

    /* write one field per line to make it easy to read back in later */
    fprintf(fp, "%d|%s|%s|%f\n", cust->id, cust->lname,
                cust->fname, cust->bal);
}

/*
 * This function allows the user to edit a customer record
 */
void edit_cust(struct cust *cust)
{
	char buf[BUFSIZ];	/* user input area */

   display_cust(cust);	/* display record as it is now */

   /* ask user for updates one field at a time */
   /* user can't edit the ID field, since that is our key */

   printf("Enter new Last Name, or ENTER to leave unchanged: ");
   gets(buf);
   if (buf[0] != '\0')
   { /* something entered - update */
   	free(cust->lname);	/* free up old field */
      cust->lname = strdup(buf);	/* update with new field */
   }

   printf("Enter new First Name, or ENTER to leave unchanged: ");
   gets(buf);
   if (buf[0] != '\0')
   { /* something entered - update */
   	free(cust->fname);	/* free up old field */
      cust->fname = strdup(buf);	/* update with new field */
   }

   printf("Enter new Balance, or ENTER to leave unchanged: ");
   gets(buf);
   if (buf[0] != '\0') cust->bal = atof(buf);
}

/*
 * This function finds a customer record matching the given ID
 *
 * If no match is found, the function returns -1
 *
 * This is basicall the "find first" algorithm from an earlier lecture
 *
 */
int find_cust(struct index ndx[], int id)
{
	int i;

   for (i=0; ndx[i].id > 0; i++)
   {
   	if (ndx[i].id == id) return i;	/* match found */
   }
   /* no match found */
   return -1;
}


/*
 * This function allows the user to enter a new cust rec
 */
struct cust *get_cust(void)
{
	char buf[BUFSIZ];
   struct cust *cust;	/* record to return */
   int id;

   printf("Enter Customer ID, or ENTER to cancel add: ");
   gets(buf);
   id = atoi(buf);
   if (id <= 0)
   {
   	printf("Customer ID must be a positive integer.  Cancelled\n");
      return NULL;
   }

	cust = alloc_cust();	/* allocate memory for cust rec */
   if (cust == NULL) return NULL;	/* error */
   cust->id = id;	/* customer ID */

   printf("Enter Last Name: ");
   gets(buf);
	cust->lname = strdup(buf);

   printf("Enter First Name: ");
   gets(buf);
   cust->fname = strdup(buf);

   printf("Enter Customer Balance: ");
   gets(buf);
   cust->bal = atof(buf);

   /* all done - return the new record */
   return cust;
}

/*
 * List all the customers
 */
void list_cust(struct index *ndx, FILE *fp)
{
	int i;
   struct cust *cust;	/* temp record */
   char buf[BUFSIZ];

   for (i=0; i < MAXENT && ndx[i].id > 0; i++)
   { /* loop through index array */
   	if (i%20 == 19)
      { /* full screen - pause for user */
      	printf("Press ENTER to continue: ");
         gets(buf);
         if (buf[0] == 'q') return;
      }
   	  cust = read_cust(ndx+i, fp);
      printf("Location %4ld: ", ndx[i].loc);
      display_cust(cust);
      free_cust(cust);  /* free up memory */
   }
}

/*
 * This function searches the database for all customers with the
 * given last name
 *
 * Since the lname field is not indexed (like the ID field), we
 * have to read each of the records from the file
 *
 * You won't notice much difference for small databases, but if
 * you had a lot of customers, this method will be much slower
 * than searching by the ID field
 */
void search_lname(struct index *ndx, char *lname, FILE *fp)
{
    int i;
    struct cust *cust;

    for (i=0; ndx[i].id > 0; i++)
    { /* loop through customer records comparing lname */
        cust = read_cust(ndx+i, fp);
        if (strcmp(cust->lname, lname) == 0) display_cust(cust);
        free_cust(cust);    /* make sure we free up memory */
    }
}


/*
 * This function creates a new customer file that contains
 * only valid records
 *
 * Any customer records that were edited are still in the file
 * New records were appended to the end of the file,
 * leaving "holes" in the file containing old records
 */
void compress_file(struct index *ndx, FILE *fp)
{
    FILE *newfp;        /* new customer file */
    struct cust *cust;  /* work record */
    int i;              /* loop variable */

    if ((newfp=fopen("cust.new", "w")) == NULL)
    {
        fprintf(stderr, "Can't open new customer file\n");
        return;
    }

    /*
     * now loop* through index reading and writing records
     * read records from old customer file and write them
     * to the new customer file
     *
     * show user what we are doing as we go
     */
     for (i=0; ndx[i].id > 0; i++)
     { /* loop until end of table */
        cust = read_cust(ndx+i, fp);    /* read old record */
        printf("Copying from Location %ld: ", ndx[i].loc);
        display_cust(cust);
        write_cust(ndx+i, cust, newfp);        /* write to new file */
        printf("New location %ld\n", ndx[i].loc);
        free_cust(cust);    /* free up memory */
     }
     /* all done - close both files */
     if (fclose(fp)) perror("Can't close old file");
     if (fclose(newfp)) perror("Can't close new file");

     /* rename files to make cust.new cust.txt */
     /*
     * the rename function isn't available in Unix
     * The way you rename a file in Unix is:
     * 	link("cust.txt", "cust.old"); // link cust.old to cust.txt
     *	unlink("cust.txt");	// delete cust.txt
     * However, the link command isn't available under Microsoft
     */
     unlink("cust.old");	/* delete old file if any */
     /* rename cust.txt to cust.old */
     if (rename("cust.txt", "cust.old")) perror("Can't rename txt to old");
     /* rename cust.new cust.txt */
     if (rename("cust.new", "cust.txt")) perror("Can't rename new to txt");
}

void do_help(void)
{
	printf("\nThis program allows you to edit a customer database\n");
   printf("The database consists of a text file containing custmer records\n");
   printf("And an index file that is used to extract records from the file\n");
   printf("All new and updated records are written to the end of the file\n");
   printf("Old records stay in the file, forming 'holes' in the file\n");
   printf("You may remove the holes with the 'compress' option\n");

}

